# (c) Copyright 2007-2015, Synapse Wireless, Inc.
"""
Functions to encode and decode SNAP RPC calls.
There are three types:
    Unicast RPC AKA "RPC"
    Multicast RPC
    Directed Multicast RPC
"""


import logging
log = logging.getLogger("snaplib.%s" % (__name__))

import datetime, binascii
from snappy import SnapConstantsHelper
from snappy.SnapConstants import *
from types import *
from snappy import BuiltIn
import sequence
import crcUtil
if __debug__:
    import timeit


SUPPORTED_PYTON_TYPES = (type(str()), type(int()), type(bool()), type(None))

# Override these default settings if you want the optional behavior
appendCrc = False # Note that this is now the default value but can be overwritten per RPC by changing it in the RpcDescriptor
validateCrc = False
promiscuousCrc = False # Probably only want to enable this if you are a sniffer
validateCrcIfPresent = False

def convert_group_if_int(group):
    if isinstance(group, int):
        if group < 0:
            #Large numbers in SNAPpy appear as negative numbers to SNAP Connect
            #This will handle the case where we're passing one of those numbers directly down
            group += 0x10000
        group = binascii.unhexlify("%04x" % group)
    assert len(group) == 2
    return group

def create_mcast_descriptor(group, ttl, func_name, *args):
    group = convert_group_if_int(group)
    descriptor = RpcDescriptor()
    descriptor.isMulticast = True
    descriptor.ttl = ttl
    descriptor.funcName = func_name
    descriptor.dstAddr = descriptor.final_dst_addr = group
    descriptor.args = args
    return descriptor


class WrongArgCount(Exception):
    pass


class RpcError(Exception):
    def __init__(self, descr):
        self.descr = descr


class RpcFunc(object):
    """Wrapper for function names, providing capability to call CFuncs by index."""
    def __init__(self, funcName):
        self.func_name = funcName
        if BuiltIn.__cfuncs__.has_key(funcName):
            # Index into the C Function Wrapper table
            idx = BuiltIn.__cfuncs__[funcName]
            self.code = "%c%c%c" % (TC_CFUNC, (idx>>8) & 0xFF, idx & 0xFF)
        else:
            self.code = None

    def __str__(self): # pragma: no cover
        return self.func_name


class RpcDescriptor(object):
    # NOTE - This object defaults to being a (unicast) RPC
    # NOTE - isDirectedMulticast is considered a special case of isMulticast
    # IOW, the only allowed combinations of isMulticast, isDirectedMulticast
    # are (False, False), (True, False), and (True, True)
    # If I were writing this code from scratch TODAY I would probably use a
    # single "type" variable, however the software requirements evolved over
    # time, and I am maintaining backwards compatibility.
    def __init__(self, expectsResponse=False, crypto=None):
        self._srcAddr = None # 3-byte string
        self.dstAddr = None    # 3-byte string or 2-byte multicast group
        self.funcName = None
        self._args = ()
        self.isMulticast = False
        self.ttl = 0 # <- Multicast or Directed Multicast Packets Only
        self.expectsResponse = expectsResponse
        self.final_dst_addr = None # <- Unicast Packets only
        self.seq = None
        self.last_hop_addr = None
        self.recv_crc = None
        self.computed_crc = None
        self.crypto = crypto
        self.appendCrc = appendCrc
        # Portal 2.6 Directed Multicast additions
        self.isDirectedMulticast = False # <- Only valid when isMulticast == True
        # These next three fields are only valid when isDirectedMulticast == True
        self.dmcast_original_ttl = 0
        self.dmcast_delay_factor = 0
        self.dmcast_dstAddrs = ''

        if __debug__:
            self.createTime = timeit.default_timer()
            self.txTime = 0
            self.sendTime = 0
            self.fn, self.lno, self.func = log.findCaller()

    def get_args(self):
        return self._args

    def set_args(self, args):
        try:
            test = iter(args)
        except TypeError:
            raise RpcError("Do not receive a list of arguments")

        for arg in args:
            if type(arg) not in SUPPORTED_PYTON_TYPES and not isinstance(arg, RpcFunc):
                raise TypeError("Unsupported type %s" % type(arg))
            elif isinstance(arg, basestring) and len(arg) > 255:
                raise TypeError("String length is greater than 255")
            elif isinstance(arg, int) and (arg < -32768 or arg > 0xFFFF):
                raise TypeError("Integer is out of range")

        self._args = args

    args = property(get_args, set_args)

    def set_srcAddr(self, addr):
        if len(addr) != SNAP2_ADDR_LEN:
            raise RpcError("Source Address must be %i bytes" % (SNAP2_ADDR_LEN))

        intAddr = int(binascii.hexlify(addr), 16)
        if intAddr >= 0xFFFFFF or intAddr <= 0x000000:
            raise RpcError("Source Address must be between \xFF\xFF\xFF and \x00\x00\x00")
        self._srcAddr = addr

    def get_srcAddr(self):
        return self._srcAddr

    srcAddr = property(get_srcAddr, set_srcAddr)

    def __call__(self, callsink):
        assert callsink
        callsink.remoteAddr = self.srcAddr
        callsink.portalNetAddr = self.dstAddr
        callsink.rpcDescriptor = self
        func = getattr(callsink, self.funcName)
        if callable(func):
            #if __debug__:
                #log.debug("Function %s called on device at %s from %s" % (rpcDescriptor.funcName, \
                                                                                                                            #binascii.hexlify(rpcDescriptor.dstAddr), \
                                                                                                                            #binascii.hexlify(rpcDescriptor.srcAddr))
            if (isinstance(func, FunctionType) and
                len(self.args) != func.func_code.co_argcount):
                raise WrongArgCount
            func(*self.args)
        elif __debug__: # pragma: no cover
            log.error("RpcDescriptor did NOT get a callable object when called")

    def __str__(self): # pragma: no cover
        addr = self.final_dst_addr
        if addr is None:
            addr = self.dstAddr
        if addr is None:
            return "No addr"
        return "%s.%s%s" % (binascii.hexlify(addr), self.funcName, self.args)

    def OnSend(self, sucessful): # pragma: no cover
        if __debug__ and sucessful:
            self.sendTime = timeit.default_timer()
            pass

    def get_encoded_data(self):
        return RpcCodec.encode(self, crypto=self.crypto)

    encoded_data = property(get_encoded_data)


class RpcCodec(object):
    @staticmethod
    def encode(rpcDescriptor, crypto=None):
        """Encode an RPC function and parameters for output to a SNAP device"""
        if __debug__: # pragma: no cover
            #Used to help wing's autocomplete
            isinstance(rpcDescriptor, RpcDescriptor)

        # Argument data starts after function and arguments (which are 3-byte PyObjects)
        argDataOffset = 3 * (len(rpcDescriptor.args) + 1)
        argData = []

        if rpcDescriptor.srcAddr is None:
            raise RpcError("No source address was set")
        elif rpcDescriptor.dstAddr is None:
            raise RpcError("No destination address/group was set")

        # Default to the Unicast RPC case (will override these as needed below)
        ttl = '' # Multicast or Directed Multicast only
        # These next four are Directed Multicast only
        original_ttl = ''
        delay_factor = ''
        target_addrs_len = ''
        target_addrs = ''

        # Encode packet header
        if rpcDescriptor.isMulticast:
            if rpcDescriptor.isDirectedMulticast:
                pid = PID_DIRECTED_MULTICAST_RPC
                # Sanity check additional fields unique to Directed Multicast
                if (len(rpcDescriptor.dmcast_dstAddrs) % 3) != 0:
                    raise RpcError("dmcast_dstAddrs length must be a multiple of 3 (including 0)")
                elif rpcDescriptor.dmcast_original_ttl <= 0 or rpcDescriptor.dmcast_original_ttl > 255:
                    raise RpcError('Original TTL for multicast must be greater than 0 and less than 256')
                elif rpcDescriptor.dmcast_delay_factor < 0 or rpcDescriptor.dmcast_delay_factor > 255:
                    raise RpcError('Delay Factor for directed multicast must be between 0 and 255')

                # 'Stringify' the additional numeric fields
                original_ttl = "%c" % rpcDescriptor.dmcast_original_ttl
                delay_factor = "%c" % rpcDescriptor.dmcast_delay_factor
                target_addrs_len = "%c" % len(rpcDescriptor.dmcast_dstAddrs)
                target_addrs = rpcDescriptor.dmcast_dstAddrs
            else:
                pid = PID_MULTICAST_RPC

            # Notice that many of the header fields are COMMON to both
            # Directed Multicast and the original Multicast packets
            if len(rpcDescriptor.dstAddr) != 2:
                raise RpcError("The multicast group must be 2 bytes")

            if rpcDescriptor.dstAddr == "\x00\x00":
                raise RpcError("Destination multicast group cannot be all zeros")
            elif rpcDescriptor.ttl <= 0 or rpcDescriptor.ttl > 255:
                raise RpcError('TTL for multicast must be greater than 0 and less than 256')
            else:
                ttl = "%c" % rpcDescriptor.ttl
            finalDstAddr = ''

        else: # Unicast RPC

            if rpcDescriptor.isDirectedMulticast:
                raise RpcError("Packet cannot be a Directed Multicast without being a Multicast Packet first")

            pid = PID_RPC
            finalDstAddr = rpcDescriptor.final_dst_addr
            if finalDstAddr is None:
                finalDstAddr = rpcDescriptor.dstAddr
            if len(finalDstAddr) != 3:
                raise RpcError("The destination address must be 3 bytes")

        if rpcDescriptor.last_hop_addr is None:
            rpcDescriptor.last_hop_addr = rpcDescriptor.srcAddr

        if rpcDescriptor.seq is None:
            seq_num = sequence.get_number(rpcDescriptor.srcAddr)
        else:
            seq_num = rpcDescriptor.seq

        if crypto and crypto.has_valid_key:
            pid |= 0x80

        # In the following, be aware that fields that "do not apply" have been set to EMPTY STRINGS ('')
        #                                                                                                            <<<<<<<<<<<<<<<<< Directed Multicast Only >>>>>>>>>>>>>>>>> 
        # PID | DstAddr (or mcast: Group and TTL) | SrcAddr | Sequence | Final DstAddr(not in mcast) | Orig SrcAddr | Orig TTL | Delay Factor | Target Addrs LEN | Target Addrs | nArgs
        #  %c |       %s            %s            |   %s    |    %c    |            %s               |      %s      |   %s     |      %s      |       %s         |       %s     | %c
        rpcCode = "%c%s%s%s%c%s%s%s%s%s%s%c" % (pid, rpcDescriptor.dstAddr, ttl, rpcDescriptor.last_hop_addr, seq_num, finalDstAddr, rpcDescriptor.srcAddr,
                                                original_ttl, delay_factor, target_addrs_len, target_addrs,
                                                len(rpcDescriptor.args))

        payloadStartIndex = len(rpcCode)-1
        if rpcDescriptor.isDirectedMulticast:
            # For Directed Multicast we extended the CRC to cover the "new" header fields too
            payloadStartIndex -= (3+len(target_addrs))

        # Encode function and argument objects
        objs = (RpcFunc(rpcDescriptor.funcName),) + rpcDescriptor.args

        # Make new "callback" function in SNAPpy more useful, by replacing the function "string" with a function "number"
        # This allows callback to be invoked on built-ins, even with no SNAPpy script loaded into the unit
        if rpcDescriptor.funcName == "callback":
            if len(objs) >= 3: # if there are ENOUGH parameters
                if isinstance( objs[2], StringType): # if the one we are interested in is the correct type
                    tempRpcFunc = RpcFunc(objs[2]) # try and make a RpcFunc out of it
                    if tempRpcFunc.code != None: # if it worked...
                        objs = objs[0:2] + (tempRpcFunc,) + objs[3:] # make the substitution. Since tuples are immutable, we must rebuild it
        #elif (rpcDescriptor.funcName == 'callout' and
                    #len(objs) >= 4 and
                    #isinstance( objs[3], StringType)):
            #tempRpcFunc = RpcFunc(objs[3]) # try and make a RpcFunc out of it
            #if tempRpcFunc.code != None: # if it worked...
                #objs = objs[0:3] + (tempRpcFunc,) + objs[4:] # make the substitution. Since tuples are immutable, we must rebuild it

        # NOTE!!! If adding args checking also add up above in RpcDescriptor
        for obj in objs:
            if isinstance(obj, NoneType):
                rpcCode += "%c\x00\x00" % TC_NONE
            elif isinstance(obj, BooleanType):   # Subclass of Int, so must be checked before IntType!
                rpcCode += "%c%c%c" % (TC_BOOL, 0, int(obj))
            elif isinstance(obj, IntType):
                # 16bit signed integer
                if -32768 <= obj <= 0xFFFF:    # Allow unsigned vals
                    rpcCode += "%c%c%c" % (TC_INT, (obj>>8) & 0xFF, obj & 0xFF)
                else:
                    raise RpcError('Int out of range')
            elif isinstance(obj, RpcFunc) and obj.code:
                rpcCode += obj.code
            elif isinstance(obj, StringType) or isinstance(obj, FunctionType) or isinstance(obj, RpcFunc):
                if isinstance(obj, StringType):
                    obj = str(obj) #Remove any unicode stuff
                # Encode functions as strings
                if isinstance(obj, FunctionType) or isinstance(obj, RpcFunc):
                    obj = obj.func_name
                    if obj is None:
                        raise RpcError("Function name cannot be None")

                # 8-bit length, followed by up to 255 byte string
                if len(obj) > 255:
                    raise RpcError('Max string length is 255')
                else:
                    strOfs = argDataOffset
                    rpcCode += "%c%c%c" % (TC_STRING, (strOfs>>8) & 0xFF, strOfs & 0xFF)
                    # Encode P-String
                    argDataOffset += len(obj) + 1
                    argData.append('%c%s' % (len(obj), obj))
            else:
                raise RpcError('Unsupported DataType: %s' % type(obj))

        pkt = rpcCode + ''.join(argData) # Add args

        if rpcDescriptor.appendCrc:
            payload = pkt[payloadStartIndex:]
            crc16 = crcUtil.getSnapCrc(payload)
            pkt = "%s%c%c" % (pkt, crc16 >> 8, crc16 & 0xff)

        try:
            pkt = "%c%s" % (len(pkt), pkt) # Prepend length
        except OverflowError:
            raise RpcError("Packet is greater than 255 bytes")

        if len(pkt) > SnapConstantsHelper.getMaxPacketSize():
            raise RpcError('Function/Args too large to encode (%d bytes, max is %d)' % (len(pkt), SnapConstantsHelper.getMaxPacketSize()))

        if crypto and crypto.has_valid_key:
            if rpcDescriptor.isMulticast: # Multicast and Directed Multicast are encrypted identically
                pkt = pkt[:MULTI_PKT_DATA] + crypto.encrypt(seq_num, pkt[MULTI_PKT_DATA:])
            else:
                pkt = pkt[:FWD_UNI_PKT_DATA] + crypto.encrypt(seq_num, pkt[FWD_UNI_PKT_DATA:])

        return pkt

    @staticmethod
    def decode(rpcString, checkLength=True, crypto=None):
        """Decode an RPC function call from a SNAP device"""
        if len(rpcString) < MRPC_FUNC_OFS:
            log.debug("Received too short of a packet to decode")
            return

        rpcDesc = RpcDescriptor(crypto=crypto)
        rpcCode = map(ord,rpcString)
        pid = rpcCode[RPC_PID_OFS]
        rpcDesc.seq = rpcCode[RPC_SEQ_OFS]

        # TODO: Actual length check for encrypted packets?
        if (checkLength and
            rpcString[RPC_LENGTH_OFS] != '\x00' and
            ord(rpcString[RPC_LENGTH_OFS]) != (len(rpcString)-1) and
            not (pid & PID_ENCRYPTION_MASK)):
            #len bytes don't include self
            log.debug("Received RPC packet with length %i not the same as reported %i" % (len(rpcString)-1, int(binascii.hexlify(rpcString[RPC_LENGTH_OFS]), 16)))
            return

        if crypto and crypto.has_valid_key:
            if pid == (PID_RPC | 0x80):
                rpcString = rpcString[:FWD_UNI_PKT_DATA] + crypto.decrypt(rpcDesc.seq, rpcString[FWD_UNI_PKT_DATA:])
            elif pid == (PID_MULTICAST_RPC | 0x80):
                rpcString = rpcString[:MULTI_PKT_DATA] + crypto.decrypt(rpcDesc.seq, rpcString[MULTI_PKT_DATA:])
            elif pid == (PID_DIRECTED_MULTICAST_RPC | 0x80):
                rpcString = rpcString[:MULTI_PKT_DATA] + crypto.decrypt(rpcDesc.seq, rpcString[MULTI_PKT_DATA:])
            rpcCode = map(ord,rpcString)

        rpcDesc.last_hop_addr = rpcString[RPC_SRCADDR_OFS : RPC_SRCADDR_OFS + SNAP2_ADDR_LEN]
        stripped_pid = pid & 0x7F # Simplifies upcoming comparisions
        if stripped_pid == PID_RPC:
            rpcDesc.final_dst_addr = rpcString[RPC_DSTADDR_OFS : RPC_DSTADDR_OFS + SNAP2_ADDR_LEN]
            rpcDesc.dstAddr = rpcString[RPC_HOPADDR_OFS : RPC_HOPADDR_OFS + SNAP2_ADDR_LEN]
            nargsOffset = RPC_NARGS_OFS
            funcOffset = RPC_FUNC_OFS
            rpcDesc.srcAddr = rpcString[RPC_ORIG_SRC_OFS : RPC_ORIG_SRC_OFS + SNAP2_ADDR_LEN]
        elif stripped_pid == PID_MULTICAST_RPC or stripped_pid == PID_DIRECTED_MULTICAST_RPC:
            rpcDesc.dstAddr = rpcString[RPC_HOPADDR_OFS : RPC_HOPADDR_OFS + 2]
            rpcDesc.ttl = int(binascii.hexlify(rpcString[RPC_TTL_OFS : RPC_TTL_OFS + 1]), 16)
            if stripped_pid == PID_MULTICAST_RPC:
                nargsOffset = MRPC_NARGS_OFS
                funcOffset = MRPC_FUNC_OFS
            else:
                rpcDesc.isDirectedMulticast = True
                rpcDesc.dmcast_original_ttl = ord(rpcString[DMRPC_ORIG_TTL_OFS])
                rpcDesc.dmcast_delay_factor = ord(rpcString[DMRPC_DELAY_FACTOR_OFS])
                target_addrs_len = ord(rpcString[DMRPC_DSTADDRS_LEN_OFS])
                rpcDesc.dmcast_dstAddrs = rpcString[DMRPC_DSTADDRS_OFS:DMRPC_DSTADDRS_OFS+target_addrs_len]
                nargsOffset = DMRPC_NARGS_BASE + target_addrs_len
                funcOffset = DMRPC_FUNC_BASE + target_addrs_len
            try:
                rpcDesc.srcAddr = rpcString[MRPC_ORIG_SRC_OFS : MRPC_ORIG_SRC_OFS + SNAP2_ADDR_LEN]
            except RpcError:
                log.debug("Invalid source address")
                return None
            rpcDesc.isMulticast = True
        else:
            return None

        # CRC computation with optional validation
        appendedCrc = None
        computedCrc = None
        if rpcDesc.isDirectedMulticast == False:
            payload = rpcString[nargsOffset:rpcCode[RPC_LENGTH_OFS]+1] # Make use of the LEN byte to ignore AES-128 padding
        else: # For Directed RPC we chose to include the NEW fields into the CRC
            payload = rpcString[DMRPC_ORIG_TTL_OFS:rpcCode[RPC_LENGTH_OFS]+1] # Make use of the LEN byte to ignore AES-128 padding
        if len(payload) < 3:
            if validateCrc:
                log.debug("Received RPC packet with too short a payload to have a CRC *and* CRC validation is currently active, packet discarded")
                return
        else: # There is enough payload present that there COULD be a trailing CRC at the end of it
            appendedCrc = (ord(payload[-2]) << 8) + ord(payload[-1])
            payload = payload[0:len(payload)-2] # stripping off the trailing CRC
            if validateCrc or validateCrcIfPresent:
                computedCrc = crcUtil.getSnapCrc(payload)
                if validateCrc and computedCrc != appendedCrc:
                    log.debug("Received RPC payload with invalid appended CRC *and* CRC validation is currently active, packet discarded")
                    return
                # validateCrcIfPresent is at the end since we have to do a little more work to see if the CRC is really there

        nArgs = rpcCode[nargsOffset]

        argDataOffset = funcOffset + 3 * (nArgs + 1)

        if argDataOffset > len(rpcCode):
            log.debug("Received RPC payload with invalid number of arguments field, packet discarded")
            return

        crc_loc = argDataOffset

        # Only string-encoded functions are accepted
        if rpcCode[funcOffset] not in [TC_STRING, TC_CFUNC]:
            return None

        argList = []

        # Traverse func and args, building the argList
        for i in range(funcOffset, funcOffset + (nArgs+1)*3, 3):
            typeCode = rpcCode[i]
            assert (i+2) < len(rpcCode) #Have gotten a list index out of range, want to catch
            intVal = rpcCode[i+1] << 8 | rpcCode[i+2]
            intVal = ((intVal + 32768) % 65536) - 32768 #Convert to signed int
            if typeCode == TC_NONE:
                argList.append(None)
            elif typeCode == TC_BOOL:
                argList.append(bool(intVal))
            elif typeCode == TC_INT:
                argList.append(intVal)
            elif typeCode == TC_STRING:
                strOfs = funcOffset + intVal
                # RPC_CRCs can catch this sort of garbage, but not everyone enables them
                # So, being more robust here too
                if (strOfs < crc_loc) or (strOfs >= len(rpcCode)):
                    log.debug("Received RPC payload with invalid string offset, packet discarded")
                    return
                strLen = rpcCode[strOfs]
                # Robustness, cont.
                if ((strOfs + 1) > len(rpcString)) or ((strOfs + 1 + strLen) > len(rpcString)):
                    log.debug("Received RPC payload with invalid string length, packet discarded")
                    return
                strVal = str(rpcString[strOfs + 1 : strOfs + 1 + strLen ])
                argDataOffset += strLen + 1
                argList.append(strVal)
                crc_loc += strLen+1
            elif typeCode == TC_CFUNC:
                for (func_name, cfunc_int) in BuiltIn.__cfuncs__.iteritems():
                    if cfunc_int == intVal:
                        argList.append(RpcFunc(func_name))
                        break
            elif __debug__: # pragma: no cover
                log.error("Unhandled type code %i" % (typeCode))

        if len(argList) == 0:
            log.error("Malformed RPC packet")
            return
        rpcDesc.funcName = argList.pop(0)
        if isinstance(rpcDesc.funcName, IntType):
            for (key, value) in BuiltIn.__cfuncs__.iteritems():
                if value == rpcDesc.funcName:
                    rpcDesc.funcName = key
                    break
            else:
                rpcDesc.funcName = str(rpcDesc.funcName)
        elif isinstance(rpcDesc.funcName, RpcFunc):
            rpcDesc.funcName = rpcDesc.funcName.func_name
        elif not isinstance(rpcDesc.funcName, StringType):
            return None

        rpcDesc.args = tuple(argList)

        # If validateCrc was on, we would have bailed up above
        # The following code is for the sniffer, and tolerates packets with or without CRCs
        if crc_loc < len(rpcString)-1: # CRC takes up TWO bytes, not just one
            rpcDesc.recv_crc = appendedCrc
            rpcDesc.computed_crc = computedCrc
            if not promiscuousCrc and validateCrcIfPresent and appendedCrc != computedCrc:
                log.debug("Received RPC payload with invalid appended CRC")
                return None

        return rpcDesc


decode = RpcCodec.decode
encode = RpcCodec.encode

